import json
import logging
import re
from abc import ABC
from functools import partial
from typing import Callable, List, Literal

import rich
import rich.progress
from datasets import Dataset, DatasetDict, load_dataset

from llm_evaluation_in_reasoning.data.question import QuestionType
from llm_evaluation_in_reasoning.eval.model import EvalModel, MajorityVoteModel


class BaseBenchDataloader(ABC):
    dataset: Dataset | DatasetDict

    def __init__(self) -> None:
        self.progress_bar: rich.progress.Progress
        self.question_key: str
        self.answer_key: str
        self.question_type: QuestionType

    def __len__(self) -> int:
        return len(self.dataset)

    async def evaluate_model(
        self,
        model: EvalModel | MajorityVoteModel,
        scorer: Callable[[str | List[str], str | int, QuestionType], bool],
    ) -> tuple[List[dict], float]:
        results: List[dict] = []
        total_correct = 0
        with self.progress_bar as progress:
            task = progress.add_task("Evaluating model", total=len(self.dataset))
            for i, example in enumerate(self.dataset):
                logging.debug(f"Processing example {i}")
                try:
                    response = await model.predict(example[self.question_key])
                    is_correct = scorer(
                        response, example[self.answer_key], self.question_type
                    )
                    results.append(
                        {
                            self.question_key: example[self.question_key],
                            "response": response,
                            self.answer_key: example[self.answer_key],
                            "is_correct": is_correct,
                        }
                    )
                    if is_correct:
                        total_correct += 1
                    progress.update(task, advance=1)
                    logging.info(
                        f"Progress: {i+1}/{len(self.dataset)} - Accuracy: {total_correct/(i+1):.2%}"
                    )

                except Exception as e:
                    logging.error(f"Error processing example {i}: {str(e)}")
                    results.append(
                        {self.question_key: example[self.question_key], "error": str(e)}
                    )
            accuracy = total_correct / len(self.dataset)
            return results, accuracy


class SimpleBenchDataloader(BaseBenchDataloader):
    def __init__(self, file_path: str, progress: rich.progress.Progress) -> None:
        with open(file_path, "r") as file:
            data = json.load(file)
        self.dataset = data["eval_data"]
        self.progress_bar = progress
        self.question_key = "prompt"
        self.answer_key = "answer"
        self.question_type = QuestionType.MULTIPLE_CHOICE


class HFDataloader(BaseBenchDataloader, ABC):
    def __init__(
        self,
        path: str,
        name: str,
        split: Literal["test", "train"] = "test",
        progress: rich.progress.Progress = rich.progress.Progress(),
    ):
        self.dataset: DatasetDict = load_dataset(path=path, name=name, split=split)
        self.dataset = self.dataset["test"]
        self.progress_bar = progress
        self.question_key = "question"
        self.question_type = QuestionType.BLANK_FILL
        self.answer_key = "answer"
        extract_with_params = partial(answer2int_gsm, anwser_key=self.answer_key)
        self.dataset = self.dataset.map(extract_with_params)


def answer2int_gsm(example: Dataset, anwser_key: str) -> int:
    answer_text = example[anwser_key]
    match = re.search(r"####\s*(\d+)", answer_text)
    if match:
        example[anwser_key] = int(match.group(1))
    else:
        example[anwser_key] = -1
    return example


class GSMSymbolic(HFDataloader):
    def __init__(
        self,
        type: Literal["main", "p1", "p2"] = "main",
        split: Literal["test", "train"] = "test",
        progress: rich.progress.Progress = rich.progress.Progress(),
    ):
        self.dataset: DatasetDict = load_dataset(
            path="apple/GSM-Symbolic", name=type, split=split
        )
        self.progress_bar = progress
        self.question_key = "question"
        self.question_type = QuestionType.BLANK_FILL
        self.answer_key = "answer"
        extract_with_params: Callable = partial(
            answer2int_gsm, anwser_key=self.answer_key
        )
        self.dataset = self.dataset.map(extract_with_params)


class GSM8K(HFDataloader):
    def __init__(
        self,
        type: Literal["main", "socratic"] = "main",
        split: Literal["test", "train"] = "test",
        progress: rich.progress.Progress = rich.progress.Progress(),
    ):
        self.dataset: DatasetDict = load_dataset(
            path="openai/gsm8k", name=type, split=split
        )
        self.progress_bar = progress
        self.question_key = "question"
        self.question_type = QuestionType.BLANK_FILL
        self.answer_key = "answer"
        extract_with_params = partial(answer2int_gsm, anwser_key=self.answer_key)
        self.dataset = self.dataset.map(extract_with_params)



from enum import Enum


class QuestionType(Enum):
    MULTIPLE_CHOICE = "multiple_choices"
    BLANK_FILL = "blank_fills"
    CUSTOM = "custom"


QUESTION_TYPE_PROMPT_MAP = {
    QuestionType.MULTIPLE_CHOICE: "You are a careful and systematic reasoning expert who excels at analyzing complex problems. For each question:\n\n1. Break down the problem into smaller components\n2. Evaluate each piece of evidence objectively\n3. Consider multiple perspectives and potential outcomes\n4. Assess the probability and practicality of each option\n5. Validate your reasoning process before concluding\n\nPresent your analysis in clear, logical steps. Support your reasoning with specific examples or evidence when possible. After your step-by-step analysis, provide your final answer in the following format:\n\nFinal Answer: X\n\nwhere X is one of the letters A, B, C, D, E, or F.",
    QuestionType.BLANK_FILL: "You are a creative and intuitive reasoning expert who excels at solving abstract problems. For each question:\n\n1. Trust your instincts and initial impressions\n2. Consider the problem as a whole\n3. Think outside the box and explore unconventional ideas\n4. Use your creativity to generate innovative solutions\n5. Follow your intuition to reach a unique and insightful conclusion\n\nPresent your answer in the following format:\n\nFinal Answer: X\n\nwhere X is a numerical value or a word that best completes the sentence.",
}



import asyncio
import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Callable, List, Literal

import litellm
import rich.logging
import rich.progress
from fire import Fire

from llm_evaluation_in_reasoning.data.dataloader import (
    GSM8K,
    BaseBenchDataloader,
    GSMSymbolic,
    SimpleBenchDataloader,
)
from llm_evaluation_in_reasoning.data.question import (
    QUESTION_TYPE_PROMPT_MAP,
    QuestionType,
)
from llm_evaluation_in_reasoning.eval.model import EvalModel, MajorityVoteModel
from llm_evaluation_in_reasoning.eval.scorer import (
    eval_majority_vote,
    eval_question_once,
)

LOGGER_LEVEL_MAP = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRTICAL": logging.CRITICAL,
}

RichHander = rich.logging.RichHandler()
ProgressBar = rich.progress.Progress(
    "[progress.description]{task.description}",
    rich.progress.BarColumn(),
    "[progress.percentage]{task.percentage:>3.0f}%",
    rich.progress.TimeRemainingColumn(),
)


def load_system_prompt(
    prompt_path: str, question_type: QuestionType = QuestionType.MULTIPLE_CHOICE
) -> str:
    try:
        with open(prompt_path, "r", encoding="utf-8") as f:
            data = json.load(f)
            prompt_name = question_type.value
            logging.info(f"Loaded system prompt for {prompt_name}")
            if prompt_name not in data["prompts"]:
                logging.error(f"prompt name '{prompt_name}' invalid")
                raise KeyError(f"prompt name '{prompt_name}' invalid")
            return data["prompts"][prompt_name]
    except FileNotFoundError:
        logging.error(f"file not found: {prompt_path}")
        raise FileNotFoundError(f"file not found: {prompt_path}")
    except json.JSONDecodeError:
        logging.error(f"JSON error: {prompt_path}")
        raise ValueError(f"JSON error: {prompt_path}")


def run_benchmark(
    model_name: str = "op-qwen-2.5-0.5b",
    dataset_path: Literal["simple_bench_public.json", "GSM-Symbolic"] = "GSM-Symbolic",
    num_responses: int = 1,
    output_dir: str = "results",
    temp: float = 0.7,
    max_tokens: int = 2048,
    top_p: float = 0.95,
    max_retries: int = 3,
    logging_level: Literal["INFO", "DEBUG", "ERROR", "WARNING", "CRITICAL"] = "INFO",
    type: str = "main",
    split: Literal["train", "test"] = "test",
    custom_prompt: str | Path | None = None,
):
    """
    Run evaluation benchmark on the specified model and dataset
    with the given parameters
    params:
        model_name: str - name of the model to evaluate
        dataset_path: str - path to the dataset to evaluate on
        num_responses: int - number of responses to collect for majority vote
        output_dir: str - directory to save results
        temp: float - temperature parameter for model
        max_tokens: int - maximum tokens for model
        top_p: float - top p parameter for model
        max_retries: int - maximum retries for model
        system_prompt_path: str - path to system prompt json file
        logging_level: str - logging level
        type: str - type of GSM-Symbolic dataset
        split: str - split of GSM-Symbolic dataset
        custom_prompt: str | Path | None - custom system prompt
    """
    # config log
    if logging_level == "DEBUG":
        litellm.set_verbose = True
    logging.basicConfig(
        level=LOGGER_LEVEL_MAP[logging_level],
        format="%(asctime)s - %(levelname)s - %(message)s",
        handlers=[RichHander],
    )
    # create output directory
    Path(output_dir).mkdir(exist_ok=True)

    # load dataset
    dataset: BaseBenchDataloader
    match dataset_path:
        case "simple_bench_public.json":
            dataset = SimpleBenchDataloader(
                dataset_path,
                progress=ProgressBar,
            )
            pass
        case "GSM-Symbolic":
            dataset = GSMSymbolic(progress=ProgressBar, type=type, split=split)
        case "GSM8K":
            dataset = GSM8K(progress=ProgressBar, type=type, split=split)

    logging.info(f"Loaded {len(dataset)} examples from {dataset_path}")

    # load system prompt
    if custom_prompt is not None:
        if isinstance(custom_prompt, Path):
            custom_prompt = custom_prompt.read_text()
        elif isinstance(custom_prompt, str):
            system_prompt = custom_prompt
        system_prompt = custom_prompt
    else:
        system_prompt = QUESTION_TYPE_PROMPT_MAP[dataset.question_type]
    # initialize eval model and scorer
    model: EvalModel | MajorityVoteModel
    model = EvalModel(
        model_name=model_name,
        temp=temp,
        max_tokens=max_tokens,
        top_p=top_p,
        max_retries=max_retries,
        system_prompt=system_prompt,
    )
    scorer: Callable[[str | List[str], str | int, QuestionType], bool]
    if num_responses > 1:
        model = MajorityVoteModel(model=model, num_responses=num_responses)
        scorer = eval_majority_vote
    else:
        scorer = eval_question_once

    # run evaluation
    logging.info(f"Starting evaluation with model: {model_name}")
    results, accuracy = asyncio.run(dataset.evaluate_model(model, scorer))

    # save results
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    result_file = Path(output_dir) / f"results_{model_name}_{timestamp}.json"

    output = {
        "model_name": model_name,
        "accuracy": accuracy,
        "num_responses": num_responses,
        "parameters": {"temperature": temp, "max_tokens": max_tokens, "top_p": top_p},
        "results": results,
    }

    with open(result_file, "w") as f:
        json.dump(output, f, indent=2)

    logging.info(f"Evaluation complete - Final accuracy: {accuracy:.2%}")
    logging.info(f"Results saved to: {result_file}")


def app():
    Fire(run_benchmark)



# Modified by ashengstd on 2025.01.10
# MIT License

# Copyright (c) 2024 simple-bench

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import asyncio
import logging
import random
from abc import ABC, abstractmethod
from typing import Generic, List, Optional, TypeVar

from dotenv import load_dotenv
from litellm import acompletion
from openai import RateLimitError

load_dotenv()

# Map model names to their corresponding model IDs, see it in litellm docs: https://docs.litellm.ai/docs/providers
MODEL_MAP = {
    "gpt-4o-mini": "gpt-4o-mini",
    "claude-3-5-sonnet-20240620": "claude-3-5-sonnet-20240620",
    "gpt-4o": "gpt-4o-2024-08-06",
    "gpt-4-turbo": "gpt-4-turbo",
    "o1-preview": "o1-preview",
    "o1-mini": "o1-mini",
    "claude-3-opus-20240229": "claude-3-opus-20240229",
    "command-r-plus": "command-r-plus-08-2024",
    "gemini-1.5-pro": "gemini/gemini-1.5-pro",
    "gemini-2.0-flash": "gemini/gemini-2.0-flash-exp",
    "llama3-405b-instruct": "fireworks_ai/accounts/fireworks/models/llama-v3p1-405b-instruct",
    "claude-3-haiku": "claude-3-haiku-20240307",
    "gemini-1.5-pro-002": "gemini/gemini-1.5-pro-002",
    "mistral-large": "mistral/mistral-large-2407",
    "grok-2": "openrouter/x-ai/grok-2",
    "op-qwen-2.5-0.5b": "ollama/qwen2.5:0.5b",
    "op-qwen-2.5-7b": "ollama/qwen2.5:7b",
    "op-deepseek-v3": "openrouter/deepseek/deepseek-chat",
    "op-gemini-2.0-flash-free": "openrouter/google/gemini-2.0-flash-thinking-exp:free",
    "op-o1-preview": "openrouter/openai/o1-preview",
}

EXPONENTIAL_BASE = 2


T = TypeVar("T", str, List[str])


class BaseModel(ABC, Generic[T]):
    @abstractmethod
    async def predict(self, prompt: str) -> T:
        pass


class EvalModel(BaseModel[str]):
    def __init__(
        self,
        model_name: str,
        system_prompt: Optional[str] = None,
        temp: float = 0.7,
        max_tokens: int = 2048,
        top_p: float = 0.95,
        max_retries: int = 3,
    ):
        if model_name not in MODEL_MAP:
            raise ValueError(f"Invalid model name: {model_name}")

        self.model_name = model_name
        self.system_prompt = system_prompt
        self.temp = None if "o1" in model_name else temp
        self.max_tokens = None if "o1" in model_name else max_tokens
        self.top_p = None if "o1" in model_name else top_p
        self.max_retries = max_retries

    async def predict(self, prompt: str) -> str:
        delay = 2.0

        for i in range(self.max_retries):
            try:
                messages = []
                if self.system_prompt is not None:
                    messages.append({"role": "system", "content": self.system_prompt})
                messages.append({"role": "user", "content": prompt})
                logging.debug(f"Sending prompt to model: {prompt}")
                response = await acompletion(
                    model=MODEL_MAP[self.model_name],
                    messages=messages,
                    temperature=self.temp,
                    max_tokens=self.max_tokens,
                    top_p=self.top_p,
                    logger_fn=None,
                )

                if response.choices[0].message.content is not None:
                    return response.choices[0].message.content
                else:
                    logging.debug("No content in response" + str(response))
                    raise Exception("No content in response")

            except RateLimitError as e:
                delay *= EXPONENTIAL_BASE * (1 + random.random())
                logging.warning(
                    f"RateLimitError, retrying after {round(delay, 2)} seconds, {i+1}-th retry...",
                    e,
                )
                await asyncio.sleep(delay)
                continue
            except Exception as e:
                logging.warning(f"Error in retry {i+1}, retrying...", e)
                continue
        logging.error(f"Failed to get response after {self.max_retries} retries")
        raise Exception("Failed to get response after max retries")


class MajorityVoteModel(BaseModel[List[str]]):
    def __init__(self, model: BaseModel[str], num_responses: int = 3):
        self.model = model
        self.num_responses = num_responses

    async def predict(self, prompt: str) -> List[str]:
        tasks = [self.model.predict(prompt) for _ in range(self.num_responses)]
        return await asyncio.gather(*tasks)



# Modified by ashengstd on 2025.01.10
# MIT License

# Copyright (c) 2024 simple-bench

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from typing import List

from llm_evaluation_in_reasoning.data.question import QuestionType


def extract_answer(output: str, question_type: QuestionType) -> str | int:
    match question_type:
        case QuestionType.MULTIPLE_CHOICE:
            output = output.strip()
            match = re.search(r"Final Answer:\s*([A-F])", output, re.IGNORECASE)
        case QuestionType.BLANK_FILL:
            output = output.replace(",", " ")
            match = re.findall(r"-?\d+\.?\d*", output)[-1]
        case _:
            logging.error("Invalid question type")
            raise ValueError("Invalid question type")

    if match:
        answer = (
            match.group(1).upper()
            if question_type == QuestionType.MULTIPLE_CHOICE
            else int(float(str(match)))
        )
        logging.info(f"Answer extracted: {answer}")
        return answer

    logging.error("No answer found in model output")
    raise ValueError("No answer found in model output")


def eval_majority_vote(
    output: List[str], answer: str | int, question_type: QuestionType
) -> bool:
    model_answers = []
    for _output in output:
        try:
            model_answers.append(extract_answer(_output, question_type=question_type))
        except ValueError:
            logging.warning("no answer in this response")
            continue  # Skip this output if extraction fails

    if not model_answers:
        logging.error("Failed to extract any valid answers from model outputs")
        raise ValueError("Failed to extract any valid answers from model outputs")

    return model_answers.count(answer) > len(model_answers) / 2


def eval_question_once(
    output: str, answer: str | int, question_type: QuestionType
) -> bool:
    model_answer = extract_answer(output, question_type)
    logging.info(f"Model answer: {model_answer} | Ground truth answer: {answer}")
    return model_answer == answer
